@@ -1,5 +1,7 @@
 package Module::Build;
 
+# $Id $
+
 # This module doesn't do much of anything itself, it inherits from the
 # modules that do the real work.  The only real thing it has to do is
 # figure out which OS-specific module to pull in.  Many of the
@@ -8,7 +10,7 @@ package Module::Build;
 
 use strict;
 use vars qw($VERSION @ISA);
-$VERSION = '0.03_1';
+$VERSION = '0.04_01';
 
 # Okay, this is the brute-force method of finding out what kind of
 # platform we're on.  I don't know of a systematic way.  These values
@@ -53,26 +55,40 @@ my %OSTYPES = qw(
 		 mpeix     MPEiX
 		);
 
-eval "use Module::Build::Platform::$^O";
-if (!$@) {
-  @ISA = ("Module::Build::Platform::$^O");
-  #warn "Using Module::Build::Platform::$^O";
+use File::Spec;
+
+# We only use this once - don't waste a symbol table entry on it.
+# More importantly, don't make it an inheritable method.
+my $load = sub {
+  my $mod = shift;
+  #warn "Using $mod";
+  eval "use $mod";
+  die $@ if $@;
+  @ISA = ($mod);
+};
+
+if (grep {-e File::Spec->catfile($_, qw(Module Build Platform), $^O) . '.pm'} @INC) {
+  $load->("Module::Build::Platform::$^O");
 
 } elsif (exists $OSTYPES{$^O}) {
-  eval "use Module::Build::Platform::$OSTYPES{$^O}";
-  die $@ if $@;
-  @ISA = ("Module::Build::Platform::$OSTYPES{$^O}");
-  #warn "Using Module::Build::Platform::$OSTYPES{$^O}";
+  $load->("Module::Build::Platform::$OSTYPES{$^O}");
 
 } else {
   warn "Unknown OS type '$^O' - using default settings\n";
-  eval "use Module::Build::Platform::Default";
-  die $@ if $@;
-  @ISA = qw(Module::Build::Platform::Default);
-  #warn "Using Module::Build::Platform::Default";
-
+  $load->("Module::Build::Platform::Default");
 }
 
+sub os_type { $OSTYPES{$^O} }
+
+# XXX This doesn't work yet - it'll allow easier subclassing of Module::Build
+sub import {
+  my $pack = shift;
+  return $pack->SUPER::import unless @_;
+  my %opts = @_;
+  if ($opts{custom}) {
+    
+  }
+}
 
 1;
 __END__
@@ -87,21 +103,13 @@ Module::Build - Build and install Perl modules
  Standard process for building & installing modules:
  
    perl Build.PL
-   ./Build             # this script created by 'perl Build.PL'
+   ./Build
    ./Build test
    ./Build install
- 
- Other actions:
- 
-   ./Build clean
-   ./Build realclean
-   ./Build fakeinstall
-   ./Build dist
-   ./Build help
 
 =head1 DESCRIPTION
 
-This is a very alpha version of a new module set I've been working on,
+This is a beta version of a new module set I've been working on,
 C<Module::Build>.  It is meant to be a replacement for
 C<ExtUtils::MakeMaker>.
 
@@ -109,20 +117,24 @@ To install C<Module::Build>, and any other module that uses
 C<Module::Build> for its installation process, do the following:
 
    perl Build.PL
-   ./Build
+   ./Build             # this script is created by 'perl Build.PL'
    ./Build test
    ./Build install
 
-Other actions so far include:
+Actions defined so far include:
 
-   ./Build clean
-   ./Build realclean
-   ./Build fakeinstall
-   ./Build dist
-   ./Build help
+  build                          help        
+  clean                          install     
+  dist                           manifest    
+  distcheck                      realclean   
+  distclean                      skipcheck   
+  distdir                        test        
+  disttest                       testdb      
+  fakeinstall                                
 
-It's like the C<MakeMaker> metaphor, except that C<Build> is a
-Perl script, not a Makefile.  State is stored in a directory called
+
+It's like the C<MakeMaker> metaphor, except that C<Build> is a short
+Perl script, not a long Makefile.  State is stored in a directory called
 C<_build/>.
 
 Any customization can be done simply by subclassing C<Module::Build> and
@@ -147,18 +159,70 @@ this module in the first place was so that subclassing is possible
 (and easy), I will certainly write more docs as the interface
 stabilizes.
 
+=head2 $m = Module::Build->new(...)
+
+Creates a new Module::Build object.  Arguments to the new() method are
+listed below.  The only required argument is the C<module_name> argument.
+
 =over 4
 
-=item * Module::Build->new(...)
+=item * module_name
 
-Creates a new Module::Build object.  The C<module_name> argument is
-required, and should be a string like C<'Your::Module'>.  The
-C<module_version> argument is optional - if not explicitly provided,
-we'll look for the version string in the module specified by
+The C<module_name> argument is required, and should be a string like
+C<'Your::Module'>.  We use it for several purposes, including finding
+the version string for this distribution, and creating a
+suitably-named distribution directory.
+
+=item * module_version
+
+The C<module_version> argument is optional - if not explicitly
+provided, we'll look for the version string in the module specified by
 C<module_name>, parsing it out according to the same rules as
 C<ExtUtils::MakeMaker> and C<CPAN.pm>.
 
-=item * add_to_cleanup
+=item * module_version_from
+
+Allows you to specify an alternate file for finding the module
+version, instead of looking in the file specified by C<module_name>.
+
+=item * prereq
+
+An optional C<prereq> argument specifies any module prerequisites that
+the current module depends on.  The prerequisites are given in a hash
+reference, where the keys are the module names and the values are
+version specifiers:
+
+ prereq => {Foo::Module => '2.4',
+            Bar::Module => 0,
+            Ken::Module => '>= 1.2, != 1.5, < 2.0'},
+
+These three version specifiers have different effects.  The value
+C<'2.4'> means that B<at least> version 2.4 of C<Foo::Module> must be
+installed.  The value C<0> means that B<any> version of C<Bar::Module>
+is acceptable, even if C<Bar::Module> doesn't define a version.  The
+more verbose value C<'E<gt>= 1.2, != 1.5, E<lt> 2.0'> means that
+C<Ken::Module>'s version must be B<at least> 1.2, B<less than> 2.0,
+and B<not equal to> 1.5.  The list of criteria is separated by commas,
+and all criteria must be satisfied.
+
+=item * c_source
+
+An optional C<c_source> argument specifies a directory which contains
+C source files that the rest of the build may depend on.  Any C<.c>
+files in the directory will be compiled to object files.  The
+directory will be added to the search path during the compilation and
+linking phases of any C or XS files.
+
+=item * autosplit
+
+An optional C<autosplit> argument specifies a file which should be run
+through the C<Autosplit::autosplit()> function.  In general I don't
+consider this a great idea, and I may even go so far as to remove this
+feature later.  Let me know if I shouldn't.
+
+=back
+
+=head2 $m->add_to_cleanup
 
 A C<Module::Build> method may call C<< $self->add_to_cleanup(@files) >>
 to tell C<Module::Build> that certain files should be removed when the
@@ -168,7 +232,7 @@ static lists can get difficult to manage.  I preferred to keep the
 responsibility for registering temporary files close to the code that
 creates them.
 
-=item * resume
+=head2 Module::Build->resume
 
 You'll probably never call this method directly, it's only called from
 the auto-generated C<Build> script.  The C<new()> method is only
@@ -176,7 +240,7 @@ called once, when the user runs C<perl Build.PL>.  Thereafter, when
 the user runs C<Build test> or another action, the C<Module::Build>
 object is created using the C<resume()> method.
 
-=item * dispatch
+=head2 $m->dispatch
 
 This method is also called from the auto-generated C<Build> script.
 It parses the command-line arguments into an action and an argument
@@ -186,7 +250,16 @@ C<ACTION_foo> method.  All arguments (including everything mentioned
 in L<ACTIONS> below) are contained in the C<< $self->{args} >> hash
 reference.
 
-=back
+=head2 $m->os_type
+
+If you're subclassing Module::Build and some code needs to alter its
+behavior based on the current platform, you may only need to know
+whether you're running on Windows, Unix, MacOS, VMS, etc. and not the
+fine-grained value of Perl's C<$^O> variable.  The C<os_type()> method
+will return a string like C<Windows>, C<Unix>, C<MacOS>, C<VMS>, or
+whatever is appropriate.  If you're running on an unknown platform, it
+will return C<undef> - there shouldn't be many unknown platforms
+though.
 
 
 =head1 ACTIONS
@@ -212,6 +285,12 @@ The following build actions are provided by default.
 
 =over 4
 
+=item * help
+
+This action will simply print out a message that is meant to help you
+use the build process.  It will show you a list of available build
+actions too.
+
 =item * build
 
 This is analogous to the MakeMaker 'make' target with no arguments.
@@ -244,6 +323,19 @@ with C<.t> in a C<t/> directory.
 If you want tests to be 'verbose', i.e. show details of test execution
 rather than just summary information, pass the argument C<verbose=1>.
 
+If you want to run tests under the perl debugger, pass the argument
+C<debugger=1>.
+
+In addition, if a file called C<visual.pl> exists in the top-level
+directory, this file will be executed as a Perl script and its output
+will be shown to the user.  This is a good place to put speed tests or
+other tests that don't use the C<Test::Harness> format for output.
+
+=item * testdb
+
+This is a synonym for the 'test' action with the C<debugger=1>
+argument.
+
 =item * clean
 
 This action will clean up any files that the build process may have
@@ -281,20 +373,93 @@ This is just like the C<install> action, but it won't actually do
 anything, it will just report what it I<would> have done if you had
 actually run the C<install> action.
 
+=item * manifest
+
+This is an action intended for use by module authors, not people
+installing modules.  It will bring the F<MANIFEST> up to date with the
+files currently present in the distribution.  You may use a
+F<MANIFEST.SKIP> file to exclude certain files or directories from
+inclusion in the F<MANIFEST>.  F<MANIFEST.SKIP> should contain a bunch
+of regular expressions, one per line.  If a file in the distribution
+directory matches any of the regular expressions, it won't be included
+in the F<MANIFEST>.
+
+The following is a reasonable F<MANIFEST.SKIP> starting point, you can
+add your own stuff to it:
+
+   ^_build
+   ^Build$
+   ^blib
+   ~$
+   \.bak$
+   ^MANIFEST\.SKIP$
+   CVS
+
+See the L<distcheck> and L<skipcheck> actions if you want to find out
+what the C<manifest> action would do, without actually doing anything.
+
 =item * dist
 
 This action is helpful for module authors who want to package up their
 module for distribution through a medium like CPAN.  It will create a
-tarball and compress it using GZIP compression.
+tarball of the files listed in F<MANIFEST> and compress the tarball using
+GZIP compression.
 
-=item * help
+=item * distcheck
 
-This action will simply print out a message that is meant to help you
-use the build process.  It will show you a list of available build
-actions too.
+Reports which files are in the build directory but not in the
+F<MANIFEST> file, and vice versa. (See L<manifest> for details)
+
+=item * skipcheck
+
+Reports which files are skipped due to the entries in the
+F<MANIFEST.SKIP> file (See L<manifest> for details)
+
+=item * distclean
+
+Performs the 'realclean' action and then the 'distcheck' action.
+
+=item * distdir
+
+Creates a directory called C<$(DISTNAME)-$(VERSION)> (if that
+directory already exists, it will be removed first).  Then copies all
+the files listed in the F<MANIFEST> file to that directory.  This
+directory is what people will see when they download your distribution
+and unpack it.
+
+=item * disttest
+
+Performs the 'distdir' action, then switches into that directory and
+runs a C<perl Build.PL>, followed by the 'build' and 'test' actions in
+that directory.
 
 =back
 
+=head1 AUTOMATION
+
+One advantage of Module::Build is that since it's implemented as Perl
+methods, you can invoke these methods directly if you want to install
+a module non-interactively.  For instance, the following Perl script
+will invoke the entire build/install procedure:
+
+ my $m = new Module::Build (module_name => 'MyModule');
+ $m->dispatch('build');
+ $m->dispatch('test');
+ $m->dispatch('install');
+
+If any of these steps encounters an error, it will throw a fatal
+exception.
+
+You can also pass arguments as part of the build process:
+
+ my $m = new Module::Build (module_name => 'MyModule');
+ $m->dispatch('build');
+ $m->dispatch('test', verbose => 1);
+ $m->dispatch('install', sitelib => '/my/secret/place/');
+
+Building and installing modules in this way skips creating the
+C<Build> script.
+
 =head1 STRUCTURE
 
 Module::Build creates a class hierarchy conducive to customization.
@@ -383,16 +548,44 @@ worrying about backward compatibility.
 Finally, Perl is said to be a language for system administration.
 Could it really be the case that Perl isn't up to the task of building
 and installing software?  Absolutely not - see the C<Cons> package for
-one example, at L<http://www.dsmit.com/cons/> .
+one example, at L<"http://www.dsmit.com/cons/"> .
 
 =back
 
 Please contact me if you have any questions or ideas.
 
+=head1 TO DO
+
+Need to implement a prerequisite mechanism, similar to MakeMaker's
+C<PREREQ_PM> stuff.
+
+There will also be a subclassing mechanism that doesn't require as
+much module infrastructure to use.  Something like this:
+
+ use Module::Build subclass => <<'EOF';
+  sub ACTION_foo {
+    ... implement the 'foo' action ...
+  }
+ EOF
+
+The current method of relying on time stamps to determine whether a
+derived file is out of date isn't likely to scale well, since it
+requires tracing all dependencies backward, it runs into problems on
+NFS, and it's just generally flimsy.  It would be better to use an MD5
+signature or the like, if available.  See C<cons> for an example.
+
+The current dependency-checking for .xs files is prone to errors.  You
+can make 'widowed' files by doing C<Build>, C<perl Build.PL>, and then
+C<Build realclean>.  Should be easy to fix, but it's got me wondering
+whether the dynamic declaration of dependencies is a good idea.
+
+- make man pages and install them.
+- append to perllocal.pod
+- write .packlist in appropriate location (needed for un-install)
 
 =head1 AUTHOR
 
-Ken Williams, ken@forum.swarthmore.edu
+Ken Williams, ken@mathforum.org
 
 =head1 SEE ALSO
 
