@@ -1,5 +1,7 @@
 package Module::Build::Base;
 
+# $Id $
+
 use strict;
 use Config;
 use File::Copy ();
@@ -24,10 +26,11 @@ sub new {
 
   $self->{args} = {%Config, @_, %$args};
 
-  $self->find_version;
-  $self->write_config;
-  
   $self->check_manifest;
+  $self->check_prereq;
+  $self->find_version;
+  
+  $self->write_config;
   
   return $self;
 }
@@ -84,13 +87,12 @@ sub version_from_file {
 		      $_
 		    }; \$$2
 		   };
-      no warnings;
-      my $result = eval $eval;
-      die "Could not eval '$_' in '$file': $@" if $@;
-      return $result;
+      local $^W;
+      return scalar eval $eval;
     }
   }
-  die "Couldn't find version string in '$file'";
+  return undef;
+  #die "Couldn't find version string in '$file'";
 }
 
 sub add_to_cleanup {
@@ -164,6 +166,48 @@ sub write_config {
   }
 }
 
+sub check_prereq {
+  my $self = shift;
+  return 1 unless $self->{prereq};
+
+  my $pass = 1;
+  while (my ($modname, $spec) = each %{$self->{prereq}}) {
+
+    my $file = $self->module_name_to_file($modname);
+    unless ($file) {
+      warn "WARNING: Prerequisite $modname isn't installed\n";
+      $pass = 0;
+      next;
+    }
+
+    my $version = $self->version_from_file($file);
+    if ($spec and !$version) {
+      warn "WARNING: Couldn't find a \$VERSION in prerequisite '$file'\n";
+      $pass = 0;
+      next;
+    }
+
+    my @conditions;
+    if ($spec =~ /^\s*([\w.]+)\s*$/) { # A plain number, maybe with dots, letters, and underscores
+      @conditions = (">= $spec");
+    } else {
+      @conditions = split /\s*,\s*/, $self->{prereq}{$modname};
+    }
+
+    foreach (@conditions) {
+      if ($_ !~ /^\s*  (<=?|>=?|==|!=)  \s*  [\w.]+  \s*$/x) {
+	warn "WARNING: Invalid prerequisite condition for $modname: $_\n";
+	next;
+      }
+      unless (eval "\$version $_") {
+	warn "WARNING: $modname version $version is installed, but we need Version($modname) $_\n";
+	$pass = 0;
+      }
+    }
+  }
+  return $pass;
+}
+
 sub rm_previous_build_script {
   my $self = shift;
   if (-e $self->{build_script}) {
@@ -179,25 +223,34 @@ sub make_build_script_executable {
 sub print_build_script {
   my ($self, $fh) = @_;
   
-  my $quoted_INC = join ', ', map "'$_'", @INC;
-  my $build_dir = File::Spec->rel2abs(File::Basename::dirname($0));
   my $build_package = ref($self);
 
+  my ($config_dir, $build_script, $build_dir) = 
+    ($self->{config_dir}, $self->{build_script},
+     File::Spec->rel2abs(File::Basename::dirname($0)));  # XXX should be property of $self
+
+  my @myINC = @INC;
+  for ($config_dir, $build_script, $build_dir, @myINC) {
+    s/([\\\'])/\\$1/g;
+  }
+
+  my $quoted_INC = join ', ', map "'$_'", @myINC;
+
   print $fh <<EOF;
 $self->{args}{startperl} -w
 
 BEGIN { \@INC = ($quoted_INC) }
 
-chdir('$build_dir');
+chdir('$build_dir') or die 'Cannot chdir to $build_dir: '.\$!;
 use $build_package;
 
 my \$build = resume $build_package (
-  config_dir => '$self->{config_dir}',
-  build_script => '$self->{build_script}',
+  config_dir => '$config_dir',
+  build_script => '$build_script',
 );
 eval {\$build->dispatch};
 my \$err = \$@;
-\$build->write_cleanup;
+\$build->write_cleanup;  # Always write, even if error occurs
 die \$err if \$err;
 
 EOF
@@ -224,8 +277,8 @@ sub check_manifest {
   # could easily be re-written into a modern Perl dialect.
 
   print "Checking whether your kit is complete...\n";
-  require ExtUtils::Manifest;
-  $ExtUtils::Manifest::Quiet = 1;
+  require ExtUtils::Manifest;  # ExtUtils::Manifest is not warnings clean.
+  local ($^W, $ExtUtils::Manifest::Quiet) = (0,1);
   
   if (my @missed = ExtUtils::Manifest::manicheck()) {
     print "Warning: the following files are missing in your kit:\n";
@@ -250,12 +303,9 @@ sub dispatch {
   }
 
   my $action = "ACTION_$self->{action}";
-  if ($self->can($action)) {
-    #print ref($self),"->$action\n";
-    $self->$action;
-  } else {
-    print "No method '$action' defined.\n";
-  }
+  print("No method '$action' defined.\n"), return unless $self->can($action);
+  
+  return $self->$action;
 }
 
 sub cull_args {
@@ -274,22 +324,48 @@ sub cull_args {
   return ($action, \%args);
 }
 
+sub super_classes {
+  my ($self, $class, $seen) = @_;
+  $class ||= ref($self);
+  $seen  ||= {};
+  
+  no strict 'refs';
+  my @super = grep {not $seen->{$_}++} $class, @{ $class . '::ISA' };
+  return @super, map {$self->super_classes($_,$seen)} @super;
+}
+
 sub ACTION_help {
   my ($self) = @_;
-  # XXX the list of actions should be autogenerated somehow.
-  
+
   print <<EOF;
 
  Usage: $0 <action> arg1=value arg2=value ...
  Example: $0 test verbose=1
  
  Actions defined:
-  build                      clean
-  test                       realclean
-  dist                       install
-  help                       fakeinstall
-
 EOF
+
+  my %actions;
+  {
+    no strict 'refs';
+    
+    foreach my $class ($self->super_classes) {
+      #print "Checking $class\n";
+      foreach ( keys %{ $class . '::' } ) {
+	$actions{$1}++ if /ACTION_(\w+)/;
+      }
+    }
+  }
+
+  my @actions = sort keys %actions;
+  # Flow down columns, not across rows
+  @actions = map $actions[($_ + ($_ % 2) * @actions) / 2],  0..$#actions;
+  
+  while (my ($one, $two) = splice @actions, 0, 2) {
+    printf("  %-12s                   %-12s\n", $one, $two||'');
+  }
+
+  print "\nSee `perldoc Module::Build` for details of the individual actions.\n";
 }
 
 sub ACTION_test {
@@ -298,42 +374,74 @@ sub ACTION_test {
   
   $self->depends_on('build');
   
-  $Test::Harness::verbose = $self->{args}{verbose} || 0;
+  local $Test::Harness::switches = '-w -d' if $self->{args}{debugger};
+  local $Test::Harness::verbose = $self->{args}{verbose} || 0;
   local $ENV{TEST_VERBOSE} = $self->{args}{verbose} || 0;
-  local @INC = (File::Spec->catdir('blib','lib'), @INC);
-  
-  if (-e 'test.pl') {
-    Test::Harness::runtests('test.pl');
-  } elsif (-e 't' and -d _) {
-    my $tests = $self->rscan_dir('t', qr{\.t$});
-    Test::Harness::runtests(@$tests);
-  } else {
-    print "No tests defined.\n";
+
+  # Make sure we test the module in blib/
+  {
+    local $SIG{__WARN__} = sub {};  # shut blib.pm up
+    eval "use blib";
   }
+  die $@ if $@;
+  
+  # Find all possible tests and run them
+  my @tests;
+  push @tests, 'test.pl'                          if -e 'test.pl';
+  push @tests, @{$self->rscan_dir('t', qr{\.t$})} if -e 't' and -d _;
+  if (@tests) {
+    Test::Harness::runtests(@tests);
+  } else {
+    print("No tests defined.\n");
+  }
+
+  # This will get run and the user will see the output.  It doesn't
+  # emit Test::Harness-style output.
+  if (-e 'visual.pl') {
+    $self->run_script('visual.pl', '-Mblib');
+  }
+}
+
+sub ACTION_testdb {
+  my ($self) = @_;
+  local $self->{args}{debugger} = 1;
+  $self->depends_on('test');
 }
 
 sub ACTION_build {
   my ($self) = @_;
   
+  if ($self->{args}{c_source}) {
+    push @{$self->{include_dirs}}, $self->{args}{c_source};
+    my $files = $self->rscan_dir($self->{args}{c_source}, qr{\.(c|PL)$});
+    foreach my $file (@$files) {
+      if ($file =~ /c$/) {
+	push @{$self->{objects}}, $self->compile_c($file);
+      } elsif ($file =~ /PL/) {
+	$self->run_perl_script($file);
+      }
+    }
+  }
+
   # What more needs to be done when creating blib/ from lib/?
-  # Currently we handle .pm, .xs, and .pod files, we don't handle .PL stuff.
-  my $files = $self->rscan_dir('lib', qr{\.(pm|pod|xs)$});
+  # Currently we handle .pm, .xs, .pod, and .PL files.
+  my $files = $self->rscan_dir('lib', qr{\.(pm|pod|xs|PL)$});
   $self->lib_to_blib($files, 'blib');
   $self->add_to_cleanup('blib');
 }
 
 sub ACTION_install {
   my ($self) = @_;
-  require ExtUtils::Install;  # Grr, uses MakeMaker
+  require ExtUtils::Install;
   $self->depends_on('build');
-  ExtUtils::Install::install($self->install_map, 1, 0);
+  ExtUtils::Install::install($self->install_map('blib'), 1, 0);
 }
 
 sub ACTION_fakeinstall {
   my ($self) = @_;
-  require ExtUtils::Install;  # Grr, uses MakeMaker
+  require ExtUtils::Install;
   $self->depends_on('build');
-  ExtUtils::Install::install($self->install_map, 1, 1);
+  ExtUtils::Install::install($self->install_map('blib'), 1, 1);
 }
 
 sub ACTION_clean {
@@ -351,18 +459,79 @@ sub ACTION_realclean {
 
 sub ACTION_dist {
   my ($self) = @_;
-  (my $dist_dir = $self->{args}{module_name}) =~ s/::/-/;
-  $dist_dir .= "-$self->{args}{module_version}";
-  mkdir $dist_dir or die "Can't create '$dist_dir/': $!";
   
-  require ExtUtils::Manifest;
-  my $dist_files = ExtUtils::Manifest::maniread('MANIFEST');
-  ExtUtils::Manifest::manicopy($dist_files, $dist_dir, 'best');
+  $self->depends_on('distdir');
+  
+  my $dist_dir = $self->dist_dir;
   
   $self->make_tarball($dist_dir);
   $self->delete_filetree($dist_dir);
 }
 
+sub ACTION_distcheck {
+  my ($self) = @_;
+  
+  require ExtUtils::Manifest;
+  local $^W; # ExtUtils::Manifest is not warnings clean.
+  ExtUtils::Manifest::fullcheck();
+}
+
+sub ACTION_skipcheck {
+  my ($self) = @_;
+  
+  require ExtUtils::Manifest;
+  local $^W; # ExtUtils::Manifest is not warnings clean.
+  ExtUtils::Manifest::skipcheck();
+}
+
+sub ACTION_distclean {
+  my ($self) = @_;
+  
+  $self->depends('realclean');
+  $self->depends('distcheck');
+}
+
+sub ACTION_distdir {
+  my ($self) = @_;
+  
+  my $dist_dir = $self->dist_dir;
+  
+  require ExtUtils::Manifest;  # ExtUtils::Manifest is not warnings clean.
+  local ($^W, $ExtUtils::Manifest::Quiet) = (0,1);
+  
+  my $dist_files = ExtUtils::Manifest::maniread('MANIFEST');
+  ExtUtils::Manifest::manicopy($dist_files, $dist_dir, 'best');
+}
+
+sub ACTION_disttest {
+  my ($self) = @_;
+
+  $self->depends_on('distdir');
+
+  my $dist_dir = $self->dist_dir;
+  chdir $dist_dir or die "Cannot chdir to $dist_dir: $!";
+  # XXX could be different names for scripts
+  $self->do_system("$^X Build.PL") or die "Error executing '$^X Build.PL' in dist directory: $!";
+  $self->do_system('./Build') or die "Error executing './Build' in dist directory: $!";
+  $self->do_system('./Build test') or die "Error executing './Build test' in dist directory: $!";
+  # XXX doesn't change back to top dir
+}
+
+sub ACTION_manifest {
+  my ($self) = @_;
+  
+  require ExtUtils::Manifest;  # ExtUtils::Manifest is not warnings clean.
+  local ($^W, $ExtUtils::Manifest::Quiet) = (0,1);
+  ExtUtils::Manifest::mkmanifest();
+}
+
+sub dist_dir {
+  my ($self) = @_;
+
+  (my $dist_dir = $self->{args}{module_name}) =~ s/::/-/;
+  return "$dist_dir-$self->{args}{module_version}";
+}
+
 sub make_tarball {
   my ($self, $dir) = @_;
   
@@ -373,9 +542,11 @@ sub make_tarball {
 }
 
 sub install_map {
-  my $self = shift;
-  my $blib = File::Spec->catfile('blib','lib');
-  return {$blib => $self->{args}{sitelib},
+  my ($self, $blib) = @_;
+  my $lib  = File::Spec->catfile($blib,'lib');
+  my $arch = File::Spec->catfile($blib,'arch');
+  return {$lib  => $self->{args}{sitelib},
+	  $arch => $self->{args}{sitearch},
 	  read  => ''};  # To keep ExtUtils::Install quiet
 }
 
@@ -402,7 +573,7 @@ sub delete_filetree {
     next unless -e $_;
     print "Deleting $_\n";
     if (-d $_) {
-      File::Path::rmtree($_, 0, 1);
+      File::Path::rmtree($_, 0, 0);
     } else {
       unlink $_;
     }
@@ -413,6 +584,13 @@ sub delete_filetree {
 sub lib_to_blib {
   my ($self, $files, $to) = @_;
   
+  # Create $to/arch to keep blib.pm happy (what a load of hooie!)
+  File::Path::mkpath( File::Spec->catdir($to, 'arch') );
+
+  if ($self->{args}{autosplit}) {
+    $self->autosplit_file($self->{args}{autosplit}, $to);
+  }
+  
   foreach my $file (@$files) {
     if ($file =~ /\.p(m|od)$/) {
       # No processing needed
@@ -422,12 +600,44 @@ sub lib_to_blib {
       $self->process_xs($file);
 
     } elsif ($file =~ /\.PL$/) {
-      # XXX Run the script
+      $self->run_perl_script($file);
 
     } else {
       warn "Ignoring file '$file', unknown extension\n";
     }
   }
+
+}
+
+sub autosplit_file {
+  my ($self, $file, $to) = @_;
+  require AutoSplit;
+  my $dir = File::Spec->catdir($to, 'lib', 'auto');
+  AutoSplit::autosplit($file, $dir);
+}
+
+sub compile_c {
+  my ($self, $file) = @_;
+  my $args = $self->{args}; # For convenience
+
+  # File name, minus the suffix
+  (my $file_base = $file) =~ s/\.[^.]+$//;
+  my $obj_file = "$file_base$args->{obj_ext}";
+  return $obj_file if $self->up_to_date($file, $obj_file);
+  
+  $self->add_to_cleanup($obj_file);
+  my $coredir = File::Spec->catdir($args->{archlib}, 'CORE');
+  my $include_dirs = $self->{include_dirs} ? join ' ', map {"-I$_"} @{$self->{include_dirs}} : '';
+  $self->do_system("$args->{cc} $include_dirs -c $args->{ccflags} -I$coredir -o $obj_file $file")
+    or die "error building $args->{dlext} file from '$file'";
+
+  return $obj_file;
+}
+
+sub run_perl_script {
+  my ($self, $script, $preargs, $postargs) = @_;
+  $preargs ||= '';   $postargs ||= '';
+  return $self->do_system("$self->{args}{perl5} $preargs $script $postargs");
 }
 
 # A lot of this looks Unixy, but actually it may work fine on Windows.
@@ -440,40 +650,47 @@ sub process_xs {
   (my $file_base = $file) =~ s/\.[^.]+$//;
 
   # .xs -> .c
-  if ($self->is_newer_than($file, "$file_base.c")) {
+  unless ($self->up_to_date($file, "$file_base.c")) {
     $self->add_to_cleanup("$file_base.c");
     
     my $xsubpp  = $self->module_name_to_file('ExtUtils::xsubpp')
       or die "Can't find ExtUtils::xsubpp in INC (@INC)";
     my $typemap =  $self->module_name_to_file('ExtUtils::typemap');
     
-    $self->do_system("$args->{perl5} -I$args->{archlib} -I$args->{privlib} $xsubpp" .
-		     " -noprototypes -typemap '$typemap' $file > $file_base.c")
+    # XXX the '> $file_base.c' isn't really a post-arg, it's redirection.  Fix later.
+    $self->run_perl_script($xsubpp, "-I$args->{archlib} -I$args->{privlib}", 
+			   "-noprototypes -typemap '$typemap' $file > $file_base.c")
       or die "error building .c file from '$file'";
   }
   
   # .c -> .o
-  if ($self->is_newer_than("$file_base.c", "$file_base$args->{obj_ext}")) {
-    $self->add_to_cleanup("$file_base$args->{obj_ext}");
-    my $coredir = File::Spec->catdir($args->{archlib}, 'CORE');
-    $self->do_system("$args->{cc} -c -o $file_base$args->{obj_ext} $args->{ccflags} -I$coredir $file_base.c")
-      or die "error building $args->{dlext} file from '$file_base.c'";
+  $self->compile_c("$file_base.c");
+
+  # The .bs and .a files don't go in blib/lib/, they go in blib/arch/auto/.
+  # Unfortunately we have to pre-compute the whole path.
+  my $archdir;
+  {
+    my @dirs = File::Spec->splitdir($file_base);
+    $archdir = File::Spec->catdir('blib','arch','auto', @dirs[1..$#dirs]);
   }
   
   # .xs -> .bs
-  if ($self->is_newer_than($file, "$file_base.bs")) {
+  unless ($self->up_to_date($file, "$file_base.bs")) {
     $self->add_to_cleanup("$file_base.bs");
     require ExtUtils::Mkbootstrap;
     print "ExtUtils::Mkbootstrap::Mkbootstrap('$file_base')\n";
     ExtUtils::Mkbootstrap::Mkbootstrap($file_base);  # Original had $BSLOADLIBS - what's that?
     {open my $fh, ">> $file_base.bs"}  # touch
   }
-  $self->copy_if_modified("$file_base.bs", 'blib');
+  $self->copy_if_modified("$file_base.bs", $archdir, 1);
   
   # .o -> .(a|bundle)
-  my $lib_file = File::Spec->catfile('blib', "$file_base.$args->{dlext}");
-  if ($self->is_newer_than("$file_base$args->{obj_ext}", $lib_file)) {
-    $self->do_system("$args->{shrpenv} $args->{cc} -o $lib_file $args->{lddlflags} $file_base$args->{obj_ext}")
+  my $lib_file = File::Spec->catfile($archdir, File::Basename::basename("$file_base.$args->{dlext}"));
+  unless ($self->up_to_date("$file_base$args->{obj_ext}", $lib_file)) {
+    my $linker_flags = $args->{extra_linker_flags} || '';
+    my $objects = $self->{objects} || [];
+    $self->do_system("$args->{shrpenv} $args->{cc} $args->{lddlflags} -o $lib_file ".
+		     "$file_base$args->{obj_ext} @$objects $linker_flags")
       or die "error building $args->{obj_ext} file from '$file_base.$args->{dlext}'";
   }
 }
@@ -485,25 +702,50 @@ sub do_system {
 }
 
 sub copy_if_modified {
-  my ($self, $file, $to) = @_;
+  my ($self, $file, $to, $flatten) = @_;
 
-  my $to_path = File::Spec->catfile($to, $file);
-  if (!-e $to_path or -M $to_path > -M "$file") {
-    # Create parent directories
-    my $path = File::Basename::dirname($file);
-    File::Path::mkpath(File::Spec->catfile($to, $path), 0, 0777);
-    
-    print "$file -> $to_path\n";
-    File::Copy::copy($file, $to_path) or die "Can't copy('$file', '$to_path'): $!";
+  my $to_path;
+  if ($flatten) {
+    my $basename = File::Basename::basename($file);
+    $to_path = File::Spec->catfile($to, $basename);
+  } else {
+    $to_path = File::Spec->catfile($to, $file);
   }
+  return if -e $to_path and -M $to_path < -M $file;  # Already fresh
+  
+  # Create parent directories
+  File::Path::mkpath(File::Basename::dirname($to_path), 0, 0777);
+  
+  print "$file -> $to_path\n";
+  File::Copy::copy($file, $to_path) or die "Can't copy('$file', '$to_path'): $!";
 }
 
-sub is_newer_than {
-  my ($self, $one, $two) = @_;
-  return 1 unless -e $two;
-  return -M $one < -M $two;
+sub up_to_date {
+  my ($self, $source, $derived) = @_;
+  my @source  = ref($source)  ? @$source  : ($source);
+  my @derived = ref($derived) ? @$derived : ($derived);
+
+  return 0 if grep {not -e} @derived;
+
+  my $most_recent_source = time / (24*60*60);
+  foreach my $file (@source) {
+    $most_recent_source = -M $file if -M $file < $most_recent_source;
+  }
+  
+  foreach my $file (@derived) {
+    return 0 if -M $file > $most_recent_source;
+  }
+  return 1;
 }
 
+
+#sub is_newer_than {
+#  my ($self, $one, $two) = @_;
+#  return 1 unless -e $two;
+#  return 0 unless -e $one;
+#  return -M $one < -M $two;
+#}
+
 1;
 __END__
 
